# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Test markdown module.
"""

from __future__ import annotations

import pytest

from antsibull_docutils.markdown import (
    Context,
    DocumentContext,
    GlobalContext,
    render_as_markdown,
)
from antsibull_docutils.utils import get_document_structure

RENDER_AS_MARKDOWN_AND_STRUCTURE_DATA = [
    (
        "empty",
        "",
        "restructuredtext",
        "",
        set(),
    ),
    (
        "hello-world",
        "Hello world.",
        "restructuredtext",
        r"Hello world\.",
        set(),
    ),
    (
        "link",
        "`foo <bar>`__",
        "restructuredtext",
        "[foo](bar)",
        set(),
    ),
    (
        "list",
        r"""
A list:

* an :emphasis:`entry`
* another :strong:`one`

  + a nested list entry
  + another one
""",
        "restructuredtext",
        r"""A list\:

- an <em>entry</em>
- another <strong>one</strong>

  - a nested list entry
  - another one""",
        set(),
    ),
    (
        "enumeration",
        r"""
An enumeration:

1. an :literal:`entry`

   it has two lines.

   .. code-block::

       And some code.

2. another one

3.

4. a last one
""",
        "restructuredtext",
        r"""An enumeration\:

1. an <code>entry</code>

   it has two lines\.

   ```
   And some code.
   ```
1. another one
1.
1. a last one""",
        set(),
    ),
    (
        "test-doc",
        r"""
========
Test doc
========

.. contents::
   :local:
   :depth: 2

A test document.

.. _label:

Test section
------------

This is a test section. It has a label.

.. code-block::

    Code without language

.. This is a comment

Another code block::

    ````
    This one has no language, either.
    ````

Test subsection
^^^^^^^^^^^^^^^

Linking the `label`_.

.. code-block:: python

    def f():
        return 42

Another section
---------------

Buh!

    Some block quote.
    Another line.

A line.

.. note::

  This is a note.
  A second line!

  And a second paragraph.
""",
        "restructuredtext",
        r"""# Test doc

- [Test section](\#test\-section)

  - [Test subsection](\#test\-subsection)
- [Another section](\#another\-section)

A test document\.

<a id="label"></a>

<a id="test-section"></a>
## Test section

This is a test section\. It has a label\.

```
Code without language
```

Another code block\:

`````
````
This one has no language, either.
````
`````

<a id="test-subsection"></a>
### Test subsection

Linking the [label](\#label)\.

```python
def f():
    return 42
```

<a id="another-section"></a>
## Another section

Buh\!

> Some block quote\.
> Another line\.

A line\.

> [!NOTE]
> This is a note\.
> A second line\!
>
> And a second paragraph\.""",
        set(),
    ),
    (
        "images",
        r"""
This is an image:

.. image:: foo.png

.. image:: foo.png
   :alt: alt text

.. image:: foo.png
   :width: 50cm

.. image:: foo.png
   :height: 50em

.. image:: foo.png
   :alt: alt text
   :width: 50ex

------------

.. image:: foo.png
   :width: 10px
   :height: 20px
""",
        "restructuredtext",
        r"""This is an image\:

![](foo\.png)
![alt text](foo\.png)
<img src="foo.png" width="50cm">
<img src="foo.png" height="50em">
<img src="foo.png" alt="alt text" width="50ex">

---

<img src="foo.png" width="10px" height="20px">""",
        set(),
    ),
    (
        "table",
        r"""
This is a table:

+------------+------------+-----------+
| Header 1   | Header 2   | Header 3  |
+============+============+===========+
| body row 1 | column 2   | column 3  |
+------------+------------+-----------+
| body row 2 | Cells may span columns.|
+------------+------------+-----------+
| body row 3 | Cells may  | - Cells   |
+------------+ span rows. | - contain |
| body row 4 |            | - blocks. |
+------------+------------+-----------+
""",
        "restructuredtext",
        r"""This is a table\:

<table>
<thead>
<tr><th class="head"><p>Header 1</p></th>
<th class="head"><p>Header 2</p></th>
<th class="head"><p>Header 3</p></th>
</tr>
</thead>
<tbody>
<tr><td><p>body row 1</p></td>
<td><p>column 2</p></td>
<td><p>column 3</p></td>
</tr>
<tr><td><p>body row 2</p></td>
<td colspan="2"><p>Cells may span columns.</p></td>
</tr>
<tr><td><p>body row 3</p></td>
<td rowspan="2"><p>Cells may
span rows.</p></td>
<td rowspan="2"><ul class="simple">
<li><p>Cells</p></li>
<li><p>contain</p></li>
<li><p>blocks.</p></li>
</ul>
</td>
</tr>
<tr><td><p>body row 4</p></td>
</tr>
</tbody>
</table>""",
        set(),
    ),
    (
        "title-ref",
        "`foo`",
        "restructuredtext",
        r"""<em class="title-reference">foo</em>""",
        set(),
    ),
    (
        "simple-rst",
        r"""`This` *is* **a** `simple <This>`_ `test <https://example.com>`__.""",
        "restructuredtext",
        r"""<em class="title-reference">This</em> <em>is</em> <strong>a</strong> [simple](This) [test](https\://example\.com)\.""",
        set(),
    ),
    (
        "complex-rst",
        r"""
==========
Main title
==========

Some text.

.. sectnum::
.. contents:: Table of Contents

.. _label:
.. _label_2:

This is a subtitle
^^^^^^^^^^^^^^^^^^

Some test with ``code``.
Some *emphasis* and **bold** text.

A `link <https://ansible.com/)>`__.

A `reference <label>`_.

A list:

* Item 1.
* Item 2.

  This is still item 2.
* Item 3.
*
* Item 5 after an empty item.

An enumeration:

1. Entry 1

   More of Entry 1
2. Entry 2
3. Entry 3
   Second line of entry 3

   - Sublist
   - Another entry

     1. Subenum

        .. code:: markdown

            Some codeblock:
            ```python
            def main(argv): ...
            ```

     2. Another entry

Another subtitle
^^^^^^^^^^^^^^^^

Some code:

.. code:: python

    def main(argv):
        if argv[1] == 'help':
            print('Help!')

.. note::

  Some note.

  This note has two paragraphs.

A sub-sub-title
~~~~~~~~~~~~~~~

Something problematic: :pep:`1000000`

Some text...

  Some block quote.

    A nested block quote.

--------

Some unformatted code::

    foo bar!
    baz bam.

A sub-sub-title
~~~~~~~~~~~~~~~

The same sub-sub-title again.
""",
        "restructuredtext",
        r"""# Main title

Some text\.

## Table of Contents

- [1   This is a subtitle](\#this\-is\-a\-subtitle)
- [2   Another subtitle](\#another\-subtitle)

  - [2\.1   A sub\-sub\-title](\#a\-sub\-sub\-title)
  - [2\.2   A sub\-sub\-title](\#a\-sub\-sub\-title\-1)
<a id="label"></a>
<a id="label-2"></a>

<a id="this-is-a-subtitle"></a>
## 1   This is a subtitle

Some test with <code>code</code>\.
Some <em>emphasis</em> and <strong>bold</strong> text\.

A [link](https\://ansible\.com/\))\.

A [reference](\#label)\.

A list\:

- Item 1\.
- Item 2\.

  This is still item 2\.
- Item 3\.
-
- Item 5 after an empty item\.

An enumeration\:

1. Entry 1

   More of Entry 1
1. Entry 2
1. Entry 3
   Second line of entry 3

   - Sublist
   - Another entry

     1. Subenum

        ````markdown
        Some codeblock:
        ```python
        def main(argv): ...
        ```
        ````
     1. Another entry

<a id="another-subtitle"></a>
## 2   Another subtitle

Some code\:

```python
def main(argv):
    if argv[1] == 'help':
        print('Help!')
```

> [!NOTE]
> Some note\.
>
> This note has two paragraphs\.

<a id="a-sub-sub-title"></a>
### 2\.1   A sub\-sub\-title

Something problematic\: <a href="#system-message-1"><span class="problematic">\:pep\:\`1000000\`</span></a>

<details>
<summary><strong>ERROR/3</strong> (&lt;string&gt;, line 77)</summary>

PEP number must be a number from 0 to 9999\; \"1000000\" is invalid\.

</details>

Some text\.\.\.

> Some block quote\.
>
> > A nested block quote\.

---

Some unformatted code\:

```
foo bar!
baz bam.
```

<a id="a-sub-sub-title-1"></a>
### 2\.2   A sub\-sub\-title

The same sub\-sub\-title again\.""",
        set(),
    ),
    (
        "system-messages",
        r"""
==========
Main title
==========

Some text.

.. Trigger a system message:

.. unknown-shit::

  Something.
""",
        "restructuredtext",
        r"""# Main title

Some text\.

<details>
<summary><strong>ERROR/3</strong> (&lt;string&gt;, line 10)</summary>

Unknown directive type \"unknown\-shit\"\.

```
.. unknown-shit::

  Something.
```

</details>""",
        set(),
    ),
    (
        "image",
        r"""Image:

.. image:: image.png
    :alt: An image

Image w/o alt:

.. image:: https://example.com/image.png

Image w/o alt, with size:

.. image:: https://example.com/image.png
    :width: 100px
    :height: 50px

Image with size:

.. image:: https://example.com/image.png
    :alt: An image
    :width: 150px
""",
        "restructuredtext",
        r"""Image\:

![An image](image\.png)

Image w/o alt\:

![](https\://example\.com/image\.png)

Image w/o alt\, with size\:

<img src="https://example.com/image.png" width="100px" height="50px">

Image with size\:

<img src="https://example.com/image.png" alt="An image" width="150px">""",
        set(),
    ),
    (
        "table",
        r""".. _tables:

======
Tables
======

Regular table:

+-----+-----+
| Foo | Bar |
+=====+=====+
| A   | B   |
+-----+-----+
| C   | D   |
|     | DD  |
|     |     |
|     | DDD |
+-----+-----+

A list table:

.. list-table::
  :width: 100%
  :widths: auto
  :header-rows: 1

  * - Foo
    - Bar
  * - A
    - B
  * - C
    - D

      DD

      DDD
""",
        "restructuredtext",
        r"""# Tables

<a id="tables"></a>

Regular table\:

<table>
<thead>
<tr><th class="head"><p>Foo</p></th>
<th class="head"><p>Bar</p></th>
</tr>
</thead>
<tbody>
<tr><td><p>A</p></td>
<td><p>B</p></td>
</tr>
<tr><td><p>C</p></td>
<td><p>D
DD</p>
<p>DDD</p>
</td>
</tr>
</tbody>
</table>

A list table\:

<table style="width: 100%;">
<thead>
<tr><th class="head"><p>Foo</p></th>
<th class="head"><p>Bar</p></th>
</tr>
</thead>
<tbody>
<tr><td><p>A</p></td>
<td><p>B</p></td>
</tr>
<tr><td><p>C</p></td>
<td><p>D</p>
<p>DD</p>
<p>DDD</p>
</td>
</tr>
</tbody>
</table>""",
        set(),
    ),
    (
        "subtitle",
        # TODO: what should we do with subtitles?
        r"""
========
Test doc
========

This ends up as a subtitle
--------------------------

Bar.

Test subsection
^^^^^^^^^^^^^^^

Foo.
""",
        "restructuredtext",
        r"""# Test doc

Bar\.

<a id="test-subsection"></a>
## Test subsection

Foo\.""",
        {"subtitle"},
    ),
    (
        "bug-2",
        r"""
A block quote:

    Some block quote.
    Another line.

    Another paragraph.

.. note::

  This is a note.
  A second line!

  And a second paragraph.

.. note::

  And another note.

- A list item
- .. note::

    This is a note.
    A second line!

    And a second paragraph.

  .. note::

    And another note.
""",
        "restructuredtext",
        r"""A block quote\:

> Some block quote\.
> Another line\.
>
> Another paragraph\.

> [!NOTE]
> This is a note\.
> A second line\!
>
> And a second paragraph\.

> [!NOTE]
> And another note\.

- A list item
- > [!NOTE]
  > This is a note\.
  > A second line\!
  >
  > And a second paragraph\.

  > [!NOTE]
  > And another note\.""",
        set(),
    ),
]


@pytest.mark.parametrize(
    "title, input, input_parser, output, unsupported_class_names",
    RENDER_AS_MARKDOWN_AND_STRUCTURE_DATA,
    ids=[e[0] for e in RENDER_AS_MARKDOWN_AND_STRUCTURE_DATA],
)
def test_render_as_markdown(
    title, input, input_parser, output, unsupported_class_names
):
    result = render_as_markdown(input, parser_name=input_parser)
    print(get_document_structure(input, parser_name=input_parser).output)
    assert result.output == output
    assert result.unsupported_class_names == unsupported_class_names


def test_global_context():
    global_context = GlobalContext()
    assert global_context.register_new_fragment("foo") == "foo"
    assert global_context.register_new_fragment("foo") == "foo-1"
    assert global_context.register_new_fragment("foo") == "foo-2"
    code = r"""
====
Test
====

Some test.

.. _foo:
.. _foobar:

Foo
^^^

This is some text. `Foo`_.
"""
    expected = r"""# Test

Some test\.

<a id="foo-3"></a>
<a id="foobar"></a>

<a id="foo-1-1"></a>
## Foo

This is some text\. [Foo](\#foo\-3)\."""
    assert (
        render_as_markdown(
            code, global_context=global_context, parser_name="restructuredtext"
        ).output
        == expected
    )
    assert global_context.register_new_fragment("foo") == "foo-4"


def test_context():
    context = Context(has_top=False)
    with pytest.raises(ValueError) as exc:
        context.add_top("test")
    assert str(exc.value) == "Context has no top part"

    subcontext = Context(has_top=True)
    with pytest.raises(ValueError) as exc:
        context.append(subcontext)
    assert str(exc.value) == "Context has no top part"

    context2 = Context(has_top=True)
    context2.append(subcontext)


def test_document_context():
    global_context = GlobalContext()
    doc_context = DocumentContext(global_context)

    with pytest.raises(ValueError) as exc:
        doc_context.pop_context()
    assert str(exc.value) == "Cannot pop last element"

    previous_top = doc_context.top

    a_context = Context()
    doc_context.push_context(a_context)
    assert doc_context.top is a_context
    doc_context.pop_context()

    assert doc_context.top is previous_top



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
MarkDown utilities.
"""

from __future__ import annotations

import re

_MD_ESCAPE = re.compile(r"""([!"#$%&'()*+,:;<=>?@[\\\]^_`{|}~.-])""")


def md_escape(text: str) -> str:
    """
    Escape a text for MarkDown.
    """
    return _MD_ESCAPE.sub(r"\\\1", text)


__all__ = ("md_escape",)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Utility code for rendering.
"""

from __future__ import annotations

import io
import typing as t
from dataclasses import dataclass

from docutils.core import publish_parts
from docutils.utils import Reporter as DocutilsReporter

SupportedParser = t.Union[t.Literal["restructuredtext"], t.Literal["markdown"]]


_DOCUTILS_PUBLISH_SETTINGS = {
    "input_encoding": "unicode",
    "file_insertion_enabled": False,
    "raw_enabled": False,
    "_disable_config": True,
    "report_level": DocutilsReporter.ERROR_LEVEL,
}


def get_docutils_publish_settings(
    *,
    warnings_stream: io.IOBase | None = None,
) -> dict[str, t.Any]:
    """
    Provide docutils publish settings.
    """
    settings = _DOCUTILS_PUBLISH_SETTINGS.copy()
    settings["warning_stream"] = warnings_stream or False
    return settings


@dataclass
class RenderResult:
    """
    A rendering result.
    """

    # The output of the renderer.
    output: str

    # The set of class names found that weren't supported by this renderer.
    unsupported_class_names: set[str]

    # List of warnings emitted
    warnings: list[str]


def get_document_structure(
    source: str, /, parser_name: SupportedParser
) -> RenderResult:
    """
    Render the document as its internal docutils structure.
    """
    warnings_stream = io.StringIO()
    parts = publish_parts(
        source=source,
        parser_name=parser_name,
        settings_overrides=get_docutils_publish_settings(
            warnings_stream=warnings_stream
        ),
    )
    whole = parts["whole"]
    return RenderResult(
        whole.decode("utf-8") if isinstance(whole, bytes) else whole,
        set(),
        warnings_stream.getvalue().splitlines(),
    )


def ensure_newline_after_last_content(lines: list[str]) -> None:
    """
    Ensure that if ``lines`` is not empty, the last entry is ``""``.
    """
    if lines and lines[-1] != "":
        lines.append("")


__all__ = (
    "SupportedParser",
    "RenderResult",
    "get_docutils_publish_settings",
    "get_document_structure",
    "ensure_newline_after_last_content",
)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Code for conversion to MarkDown.
"""

from __future__ import annotations

import io
import typing as t
from urllib.parse import quote as _urllib_quote

from docutils import nodes, writers
from docutils.core import publish_parts
from docutils.writers.html5_polyglot import HTMLTranslator
from docutils.writers.html5_polyglot import Writer as HTMLWriter

from .html_utils import html_argument_escape, html_escape
from .md_utils import md_escape
from .utils import RenderResult, SupportedParser, get_docutils_publish_settings


class GlobalContext:
    """
    The global conversion context.
    """

    # How labels are mapped to link fragments
    labels: dict[str, str]

    # HTML fragments used so far
    fragments: set[str]

    def __init__(self):
        self.labels = {}
        self.fragments = set()

    def register_label(self, label: str) -> str:
        """
        Register a label. Produces the corresponding link fragment.
        """
        if label in self.labels:
            return self.labels[label]
        fragment = self.register_new_fragment(_urllib_quote(label, safe=""))
        self.labels[label] = fragment
        self.fragments.add(fragment)
        return fragment

    def register_new_fragment(self, fragment: str) -> str:
        """
        Register a new fragment.

        If it already exists, will be modified until a new one is found.
        """
        counter = 0
        appendix = ""
        while (full_fragment := f"{fragment}{appendix}") in self.fragments:
            counter += 1
            appendix = f"-{counter}"
        self.fragments.add(full_fragment)
        return full_fragment


class Context:
    """
    Base context class for MarkDown conversion.
    """

    escape: bool

    _has_top: bool

    _top: list[str]
    _main: list[str]

    def __init__(self, has_top=False, escape=True):
        self.escape = escape
        self._has_top = has_top
        self._top = []
        self._main = []

    def add_top(self, text: str) -> None:
        """
        Add text to the top part.
        """
        if not self._has_top:
            raise ValueError("Context has no top part")
        self._top.append(text)

    def add_main(self, text: str) -> None:
        """
        Add text to the main part.
        """
        self._main.append(text)

    def replace_main(self, lines: list[str] | None = None) -> None:
        """
        Replace the content of the main part by the new list of strings.
        """
        self._main.clear()
        if lines:
            self._main.extend(lines)

    def ensure_newline(self) -> None:
        """
        Ensure that the main part ends with a newline, if it is not empty.
        """
        if self._main and not self._main[-1].endswith("\n"):
            self._main.append("\n")

    def ensure_double_newline(self) -> None:
        """
        Ensure that the main part ends with two newlines, if it is not empty.
        """
        if self._main:
            last = "".join(self._main[-2:])
            if not last.endswith("\n\n"):
                if last.endswith("\n"):
                    self._main.append("\n")
                else:
                    self._main.append("\n\n")

    def append(
        self,
        context: "Context",
    ) -> None:
        """
        Append another context to this context.
        """
        if context._has_top:  # pylint: disable=protected-access
            if not self._has_top:
                raise ValueError("Context has no top part")
            self._top.extend(context._top)  # pylint: disable=protected-access
        self._main.extend(context._main)  # pylint: disable=protected-access

    def close(self) -> None:
        """
        Close the context.

        Should be called before using its value / content, for example by
        passing it to ``append()`` or calling ``get_text()``.
        """

    def get_text(self) -> str:
        """
        Return the content of the context as a string.
        """
        return "".join(self._top + self._main)


class ListContext(Context):
    """
    Context for lists (enumeration, bullet list).
    """

    first_indent: str
    next_indent: str

    def __init__(self, indent: str):
        super().__init__()
        self.first_indent = indent
        self.next_indent = " " * len(indent)


def get_language_from_literal_block(node: nodes.literal_block) -> str | None:
    """
    Obtain programming language from a literal block.
    """
    # Some versions store the language in the "language" attribute
    if "language" in node.attributes:
        return node.attributes["language"]
    # Some others simply store it as a class name that's not "code" or "code-block"
    for class_name in node.attributes["classes"]:
        if class_name not in ("code", "code-block"):
            return class_name
    return None


def count_leading_backticks(line: str) -> int:
    """
    Count the number of leading backticks of a string.
    """
    count = 0
    length = len(line)
    while count < length and line[count] == "`":
        count += 1
    return count


class DocumentContext:
    """
    Conversion context for a document.
    """

    section_depth: int
    global_context: GlobalContext
    unknown_node_types: set[str]
    created_labels: set[str]

    _contexts: list[Context]

    def __init__(self, global_context: GlobalContext):
        self.section_depth = 0
        self.global_context = global_context
        self.unknown_node_types = set()
        self.created_labels = set()
        self._contexts = [Context(has_top=True)]

    def push_context(self, context: Context) -> None:
        """
        Add a new context on the context stack.
        """
        self._contexts.append(context)

    def pop_context(self) -> None:
        """
        Pop the top-most context from the context stack.

        The last context cannot be popped, trying to do so will result in a
        ``ValueError`` exception.
        """
        if len(self._contexts) < 2:
            raise ValueError("Cannot pop last element")
        context = self._contexts.pop()
        context.close()
        self._contexts[-1].append(context)

    @property
    def top(self) -> Context:
        """
        Return the top-most context from the context stack.
        """
        return self._contexts[-1]


_Class = t.TypeVar("_Class")


def _add_unsupported_classes(clazz: _Class) -> _Class:
    def get_visit(class_name: str):
        # pylint: disable-next=unused-argument
        def visit_unsupported(self: t.Any, node: nodes.Node):
            # pylint: disable-next=protected-access
            self._context.unknown_node_types.add(class_name)
            raise nodes.SkipNode

        return visit_unsupported

    for class_name in nodes.node_class_names:
        function_name = f"visit_{class_name}"
        if not hasattr(clazz, function_name):
            setattr(clazz, function_name, get_visit(class_name))

    return clazz


def _add_simple(
    definitions: dict[str, tuple[str, str]]
) -> t.Callable[[_Class], _Class]:
    def decorator(clazz: _Class) -> _Class:
        def get_visit(start: str):
            # pylint: disable-next=unused-argument
            def visit(self: t.Any, node: nodes.Node) -> None:
                if start:
                    # pylint: disable-next=protected-access
                    self._context.top.add_main(start)

            return visit

        def get_depart(end: str):
            # pylint: disable-next=unused-argument
            def depart(self: t.Any, node: nodes.Node) -> None:
                if end:
                    # pylint: disable-next=protected-access
                    self._context.top.add_main(end)

            return depart

        for class_name, (start, end) in definitions.items():
            setattr(clazz, f"visit_{class_name}", get_visit(start))
            setattr(clazz, f"depart_{class_name}", get_depart(end))
        return clazz

    return decorator


def _add_skip(class_names: list[str]) -> t.Callable[[_Class], _Class]:
    def decorator(clazz: _Class) -> _Class:
        def visit(self: t.Any, node: nodes.Node) -> None:
            raise nodes.SkipNode

        for class_name in class_names:
            setattr(clazz, f"visit_{class_name}", visit)
        return clazz

    return decorator


@_add_unsupported_classes
@_add_simple(
    {
        "document": ("", ""),
        "inline": ("", ""),
        "generated": ("", ""),
        "literal": ("<code>", "</code>"),
        "emphasis": ("<em>", "</em>"),
        "strong": ("<strong>", "</strong>"),
        "subscript": ("<sub>", "</sub>"),
        "superscript": ("<sup>", "</sup>"),
    }
)
@_add_skip(
    [
        "comment",
    ]
)
class Translator(nodes.NodeVisitor):  # pylint: disable=too-many-public-methods
    """
    Translates a docutils node tree to MarkDown in a ``DocumentContext``.
    """

    _context: DocumentContext

    def __init__(self, document, document_context: DocumentContext):
        super().__init__(document)
        self._context = document_context

    def get_text(self) -> str:
        """
        Return the result as a text.

        Must only be called after processing the document, and at most once.
        """
        context = self._context.top
        context.close()
        return context.get_text().lstrip("\n").rstrip("\n")

    def _add_label(self, label: str) -> None:
        if label in self._context.created_labels:
            return
        self._context.created_labels.add(label)
        ref = self._context.global_context.register_label(label)
        self._context.top.add_main(f'<a id="{ref}"></a>\n')

    # Node: paragraph

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_paragraph(self, node: nodes.section) -> None:
        self._context.top.ensure_double_newline()

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_paragraph(self, node: nodes.section) -> None:
        self._context.top.ensure_double_newline()

    # Node: title

    # pylint: disable-next=missing-function-docstring
    def visit_title(self, node: nodes.title) -> None:
        level = self._context.section_depth + 1
        title = node.astext()
        text = f'{"#" * level} {md_escape(title)}\n\n'
        if self._context.section_depth == 0:
            self._context.top.add_top(text)
        else:
            self._context.top.ensure_newline()
            self._context.top.add_main(text)
        raise nodes.SkipNode

    # Node: section

    # pylint: disable-next=missing-function-docstring
    def visit_section(self, node: nodes.section) -> None:
        self._context.top.ensure_double_newline()
        if "ids" in node.attributes:
            for ref in node.attributes["ids"]:
                self._add_label(ref)
        self._context.section_depth += 1

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_section(self, node: nodes.section) -> None:
        self._context.section_depth -= 1

    # Node: text

    # pylint: disable-next=missing-function-docstring,invalid-name
    def visit_Text(self, node: nodes.Text) -> None:
        text = node.astext()
        if self._context.top.escape:
            text = md_escape(text)
        self._context.top.add_main(text)

    # pylint: disable-next=missing-function-docstring,invalid-name
    def depart_Text(self, node: nodes.Text) -> None:
        pass

    # Node: target

    # pylint: disable-next=missing-function-docstring
    def visit_target(self, node: nodes.target) -> None:
        if "refid" in node.attributes:
            self._add_label(node.attributes["refid"])

    # pylint: disable-next=missing-function-docstring
    def depart_target(self, node: nodes.target) -> None:
        pass

    # Node: reference

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_reference(self, node: nodes.reference) -> None:
        self._context.top.add_main("[")

    # pylint: disable-next=missing-function-docstring
    def depart_reference(self, node: nodes.reference) -> None:
        name: str
        ref: str
        if "refuri" in node.attributes:
            ref = node.attributes["refuri"]
            name = node.attributes.get("name") or "link"
        elif "refid" in node.attributes:
            ref = node.attributes["refid"]
            name = "reference"
        else:
            # docutils fails an assertion at this point
            raise ValueError(
                f"Do not know how to handle reference with attributes {node.attributes!r}"
            )
        if name == "reference":
            ref = self._context.global_context.register_label(ref)
            ref = f"#{ref}"
        self._context.top.add_main(f"]({md_escape(ref)})")

    # Node: bullet_list

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_bullet_list(self, node: nodes.bullet_list) -> None:
        self._context.top.ensure_double_newline()
        self._context.push_context(ListContext("- "))

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_bullet_list(self, node: nodes.bullet_list) -> None:
        self._context.pop_context()

    # Node: enumerated_list

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_enumerated_list(self, node: nodes.enumerated_list) -> None:
        self._context.top.ensure_double_newline()
        self._context.push_context(ListContext("1. "))

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_enumerated_list(self, node: nodes.enumerated_list) -> None:
        self._context.pop_context()

    # Node: list_item

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_list_item(self, node: nodes.list_item) -> None:
        list_context = t.cast(ListContext, self._context.top)

        class ListItemContext(Context):
            """
            Context for a list item.
            """

            def close(self) -> None:
                text = self.get_text().strip("\n")
                self.replace_main()
                indent = list_context.first_indent
                for i, line in enumerate(text.splitlines()):
                    if i and not line.rstrip(" "):
                        self.add_main("\n")
                    else:
                        self.add_main(f"{indent}{line}\n")
                    indent = list_context.next_indent
                if not self._main:
                    self.add_main(f"{indent.rstrip(' ')}\n")

        self._context.push_context(ListItemContext())

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_list_item(self, node: nodes.list_item) -> None:
        self._context.pop_context()

    # Node: literal_block

    # pylint: disable-next=missing-function-docstring
    def visit_literal_block(self, node: nodes.literal_block) -> None:
        self._context.top.ensure_newline()
        language = get_language_from_literal_block(node)

        class LiteralContext(Context):
            """
            Context for a literal block.
            """

            def __init__(self):
                super().__init__(escape=False)

            def close(self) -> None:
                self.ensure_newline()
                text = self.get_text()
                backticks_needed = "```"
                for line in text.splitlines():
                    if line.startswith(backticks_needed):
                        backticks_needed = "`" * (count_leading_backticks(line) + 1)
                self.replace_main(
                    [
                        f'{backticks_needed}{language or ""}\n',
                        text,
                        f"{backticks_needed}\n",
                    ]
                )

        self._context.push_context(LiteralContext())

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_literal_block(self, node: nodes.literal_block) -> None:
        self._context.pop_context()

    # Node: block_quote

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_block_quote(self, node: nodes.block_quote) -> None:
        class BlockQuoteContext(Context):
            """
            Context for a block quote.
            """

            def close(self) -> None:
                self.ensure_newline()
                lines = self.get_text().rstrip("\n").splitlines()
                self.replace_main(
                    [(f"> {line}\n" if line else ">\n") for line in lines]
                )

        self._context.push_context(BlockQuoteContext())

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_block_quote(self, node: nodes.block_quote) -> None:
        self._context.pop_context()

    # Node: system_message

    # pylint: disable-next=missing-function-docstring
    def visit_system_message(self, node: nodes.system_message) -> None:
        self._context.top.ensure_double_newline()
        self._context.top.add_main("<details>\n")
        message_type = html_escape(node["type"])
        message_level = node["level"]
        source = node["source"]
        if node.hasattr("line"):
            source = f'{source}, line {node["line"]}'
        source = html_escape(source)
        self._context.top.add_main(
            f"<summary><strong>{message_type}/{message_level}</strong> ({source})</summary>\n\n"
        )

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_system_message(self, node: nodes.system_message) -> None:
        self._context.top.ensure_double_newline()
        self._context.top.add_main("</details>\n\n")

    # Node: system_message

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_note(self, node: nodes.note) -> None:
        class NoteContext(Context):
            """
            Context for a note.
            """

            def close(self) -> None:
                self.ensure_double_newline()
                lines = self.get_text().rstrip("\n").splitlines()
                # See https://github.com/orgs/community/discussions/16925 for the syntax
                self.replace_main(
                    ["> [!NOTE]\n"]
                    + [f"> {line}\n" if line else ">\n" for line in lines]
                )

        self._context.top.ensure_double_newline()
        self._context.push_context(NoteContext())

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_note(self, node: nodes.note) -> None:
        self._context.pop_context()

    # Node: topic

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_topic(self, node: nodes.topic) -> None:
        # The main use of <topic> elements is for Table of Contents
        class TopicContext(Context):
            """
            Context for a topic.
            """

        self._context.section_depth += 1
        self._context.push_context(TopicContext())

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_topic(self, node: nodes.topic) -> None:
        self._context.pop_context()
        self._context.section_depth -= 1

    # Node: image

    # pylint: disable-next=missing-function-docstring
    def visit_image(self, node: nodes.image) -> None:
        self._context.top.ensure_newline()
        alt = node.attributes.get("alt") or ""
        uri = node.attributes["uri"]
        width = node.attributes.get("width")
        height = node.attributes.get("height")
        if width is not None or height is not None:
            # We can only handle width and height with a HTML <img> tag
            attributes = []
            if alt:
                attributes.append(f'alt="{html_argument_escape(alt)}"')
            if width:
                attributes.append(f'width="{html_argument_escape(str(width))}"')
            if height:
                attributes.append(f'height="{html_argument_escape(str(height))}"')
            self._context.top.add_main(
                f'<img src="{html_argument_escape(uri)}" {" ".join(attributes)}>\n'
            )
        else:
            self._context.top.add_main(f"![{md_escape(alt)}]({md_escape(uri)})\n")
        raise nodes.SkipNode

    # Node: transition

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_transition(self, node: nodes.transition) -> None:
        self._context.top.ensure_double_newline()
        self._context.top.add_main("---\n\n")
        raise nodes.SkipNode

    # Node: title_reference

    # pylint: disable-next=missing-function-docstring,unused-argument
    def visit_title_reference(self, node: nodes.title_reference) -> None:
        # This usually happens if someone uses single backticks instead of double backticks
        self._context.top.add_main('<em class="title-reference">')

    # pylint: disable-next=missing-function-docstring,unused-argument
    def depart_title_reference(self, node: nodes.title_reference) -> None:
        self._context.top.add_main("</em>")

    # Node: problematic

    # pylint: disable-next=missing-function-docstring
    def visit_problematic(self, node):
        if "refid" in node.attributes:
            self._context.top.add_main(
                f'<a href="#{html_argument_escape(node.attributes["refid"])}">'
            )
        self._context.top.add_main('<span class="problematic">')

    # pylint: disable-next=missing-function-docstring
    def depart_problematic(self, node):
        self._context.top.add_main("</span>")
        if "refid" in node.attributes:
            self._context.top.add_main("</a>")

    # Node: problematic

    # pylint: disable-next=missing-function-docstring
    def visit_table(self, node):
        translator = HTMLTranslator(self.document)
        node.walkabout(translator)
        self._context.top.add_main("".join(translator.body).strip())
        raise nodes.SkipNode


class MarkDownWriter(writers.Writer):
    """
    Create a MarkDown document from a docutils node tree.
    """

    # Needed to be able to run HTMLTranslator on nodes
    settings_spec = HTMLWriter.settings_spec

    def __init__(self, document_context: DocumentContext):
        writers.Writer.__init__(self)
        self.document_context = document_context
        self.translator_class = Translator

    def translate(self) -> None:
        """
        Translate the document node tree to MarkDown.
        """
        translator = self.translator_class(self.document, self.document_context)
        if self.document:
            self.document.walkabout(translator)
        self.output = translator.get_text()


def render_as_markdown(
    source: str,
    /,
    parser_name: SupportedParser,
    global_context: GlobalContext | None = None,
    source_path: str | None = None,
    destination_path: str | None = None,
) -> RenderResult:
    """
    Render the document as MarkDown.

    :arg source: The document.
    :kwarg parser_name: The parser to use for ``source``.
    :kwarg global_context: An optional ``GlobalContext`` object.
    :kwarg source_path: An optional path for ``source``, to be used in error
        messages.
    :kwarg destination_path: An optional path for where the destination will
        be stored, to be used in error messages.
    :return: A ``RenderResult`` object containing the output and further
        information.
    """
    if global_context is None:
        global_context = GlobalContext()
    document_context = DocumentContext(global_context)
    warnings_stream = io.StringIO()
    # pylint: disable=duplicate-code
    parts = publish_parts(
        source=source,
        source_path=source_path,
        destination_path=destination_path,
        parser_name=parser_name,
        writer=MarkDownWriter(document_context),
        settings_overrides=get_docutils_publish_settings(
            warnings_stream=warnings_stream
        ),
    )
    # pylint: enable=duplicate-code
    whole = parts["whole"]
    return RenderResult(
        whole.decode("utf-8") if isinstance(whole, bytes) else whole,
        document_context.unknown_node_types,
        warnings_stream.getvalue().splitlines(),
    )


__all__ = ("GlobalContext", "render_as_markdown")



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Test utils module.
"""

from __future__ import annotations

import pytest

from antsibull_docutils.utils import (
    ensure_newline_after_last_content,
    get_document_structure,
    get_docutils_publish_settings,
)


def test_get_docutils_publish_settings():
    result = get_docutils_publish_settings()
    assert result["raw_enabled"] is False
    assert result["warning_stream"] is False

    result = get_docutils_publish_settings(warnings_stream="")
    assert result["raw_enabled"] is False
    assert result["warning_stream"] is False

    result = get_docutils_publish_settings(warnings_stream="foo")
    assert result["raw_enabled"] is False
    assert result["warning_stream"] == "foo"


def test_get_document_structure():
    result = get_document_structure(
        r"""Foo Bar :anunknownrole:`foo`

.. anunknownclass:
""",
        parser_name="restructuredtext",
    )
    print(result)
    assert isinstance(result.output, str)
    assert "Foo Bar" in result.output
    assert result.unsupported_class_names == set()
    assert any("anunknownrole" in line for line in result.warnings)
    assert all("anunknownclass" not in line for line in result.warnings)


@pytest.mark.parametrize(
    "lines, expected",
    [
        (
            [],
            [],
        ),
        (
            [
                "",
            ],
            [
                "",
            ],
        ),
        (
            [
                "Test",
            ],
            [
                "Test",
                "",
            ],
        ),
        (
            [
                "Test",
                "",
            ],
            [
                "Test",
                "",
            ],
        ),
    ],
)
def test_ensure_newline_after_last_content(lines, expected):
    lines = lines.copy()
    ensure_newline_after_last_content(lines)
    assert lines == expected



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Utilities for ReStructuredText.
"""

from __future__ import annotations

from docutils.utils import column_width as _column_width


def rst_escape(value: str, /, escape_ending_whitespace=False) -> str:
    """
    Escape RST specific constructs.
    """
    value = value.replace("\\", "\\\\")
    value = value.replace("<", "\\<")
    value = value.replace(">", "\\>")
    value = value.replace("_", "\\_")
    value = value.replace("*", "\\*")
    value = value.replace("`", "\\`")

    if escape_ending_whitespace and value.endswith(" "):
        value = value + "\\ "
    if escape_ending_whitespace and value.startswith(" "):
        value = "\\ " + value

    return value


def column_width(string: str) -> int:
    """
    Return column width of string.

    This is needed for title under- and overlines.
    """
    return _column_width(string)


__all__ = ("rst_escape", "column_width")



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Test rst_utils module.
"""

from __future__ import annotations

import pytest

from antsibull_docutils.rst_utils import column_width, rst_escape


@pytest.mark.parametrize(
    "text, escape_ending_whitespace, escaped",
    [
        ("", False, ""),
        (" ", False, " "),
        (" ", True, r"\  \ "),
        (r"\<>_*`", False, r"\\\<\>\_\*\`"),
    ],
)
def test_rst_escape(text, escape_ending_whitespace, escaped):
    result = rst_escape(text, escape_ending_whitespace=escape_ending_whitespace)
    assert result == escaped


@pytest.mark.parametrize(
    "text, width",
    [
        ("", 0),
        ("a", 1),
        ("A", 1),
        ("A test", 6),
        ("学校", 4),
    ],
)
def test_column_width(text, width):
    result = column_width(text)
    assert result == width



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Test md_utils module.
"""

from __future__ import annotations

import pytest

from antsibull_docutils.md_utils import md_escape


@pytest.mark.parametrize(
    "text, escaped",
    [
        ("", ""),
        ("This is a test, ok.", r"This is a test\, ok\."),
        (
            r"""!"#$%&'()*+,:;<=>?@[\]^_`{|}~.-""",
            r"""\!\"\#\$\%\&\'\(\)\*\+\,\:\;\<\=\>\?\@\[\\\]\^\_\`\{\|\}\~\.\-""",
        ),
        ("* foo", r"\* foo"),
    ],
)
def test_md_escape(text, escaped):
    result = md_escape(text)
    print(result)
    print(escaped)
    assert result == escaped



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
MarkDown utilities.
"""

from __future__ import annotations

from html import escape as _html_escape


def html_escape(text: str) -> str:
    """
    Escape a text for HTML.
    """
    return _html_escape(text).replace("&quot;", '"')


def html_argument_escape(text: str) -> str:
    """
    Escape a text for HTML inside an argument delimited with double quotes (").
    """
    return _html_escape(text)


__all__ = ("html_escape", "html_argument_escape")



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2024, Ansible Project

"""
Test html_utils module.
"""

from __future__ import annotations

import pytest

from antsibull_docutils.html_utils import html_argument_escape, html_escape


@pytest.mark.parametrize(
    "text, escaped",
    [
        ("", ""),
        ("Te'st", "Te&#x27;st"),
        (
            '<a href="https://example.com/?foo=bar&baz=bam">Test</a>',
            '&lt;a href="https://example.com/?foo=bar&amp;baz=bam"&gt;Test&lt;/a&gt;',
        ),
    ],
)
def test_html_escape(text, escaped):
    result = html_escape(text)
    assert result == escaped


@pytest.mark.parametrize(
    "text, escaped",
    [
        ("", ""),
        ("Te'st", "Te&#x27;st"),
        (
            '<a href="https://example.com/?foo=bar&baz=bam">Test</a>',
            "&lt;a href=&quot;https://example.com/?foo=bar&amp;baz=bam&quot;&gt;Test&lt;/a&gt;",
        ),
    ],
)
def test_html_argument_escape(text, escaped):
    result = html_argument_escape(text)
    assert result == escaped
